{
  filename - CTUniHandlerClient.pas

  Purpose:
       This file contains the declarations for implementation of Pipes (a section of shared memory that processes use for communication).
}

{$I+,H+}
unit CTUniHandlerClient;
{
History (dd.mm.yy):
1.10.2010 start of development
}

interface

uses Windows;

const TXT_NACK = 'NACK';
      TXT_NACK_DELIMITER = ' ';
      TXT_NACK_ERROR_1 = '1';   // error .. communication error (wrong check sum, wrong packet)

type
  TLogProcedure = procedure(const Text: string) of object; // callback function for logging in higher-level layer

  TUniHandlerClient = class
  private
    hPipe: THandle;
    pLogProcedure: TLogProcedure;

    procedure ErrorMessageWindow;
    function CountCheckSum(const Data: array of Byte; const Index: integer; const Length: Word): Word;
    function WrapDataToPacket(var PacketData: array of Byte; const Buffer: string): Boolean;
    function UnwrapDataFromPacket(var Buffer: string; const PacketData: array of Byte): Boolean;
    procedure WriteToLog(const Text: string);

  protected
  public
    constructor Create(const pLogProcedureArg: TLogProcedure);
    destructor Destroy; override;

    function Connect: Boolean;
    function Disconnect: Boolean;
    function Write(const Buffer: string): Boolean;
    function Read(var Buffer: string): Boolean;
  end;

implementation

uses SysUtils;

const TXT_PIPE_NAME = '\\.\pipe\PG4UWMCPipeName4RemoteControl';

      PACKET_PROTECTION_SIZE = 5; // Leading byte (1B), Data length (2B), Check sum (2B)
      PACKET_SIZE = High(Word) + PACKET_PROTECTION_SIZE;
      PACKET_LEADING_BYTE = $7E;

constructor TUniHandlerClient.Create(const pLogProcedureArg: TLogProcedure);
begin
  inherited Create;

  hPipe := INVALID_HANDLE_VALUE;
  pLogProcedure := pLogProcedureArg;
end;

destructor TUniHandlerClient.Destroy;
begin
  Disconnect;  // disconnect pipe

  inherited Destroy;
end;

function TUniHandlerClient.Connect: Boolean;
begin
  Result := False;

  if not WaitNamedPipe(TXT_PIPE_NAME, NMPWAIT_USE_DEFAULT_WAIT) then
  begin
    ErrorMessageWindow;
    Exit;
  end;

  hPipe := CreateFile(TXT_PIPE_NAME, GENERIC_READ or GENERIC_WRITE, 0,
                      nil,
                      OPEN_EXISTING,
                      0,
                      0);

  if (hPipe = INVALID_HANDLE_VALUE) then
  begin
    ErrorMessageWindow;
    Exit;
  end;

  Result := True;
end;

function TUniHandlerClient.Disconnect: Boolean;
begin
  Result := False;

  if hPipe <> INVALID_HANDLE_VALUE then
  begin
    if CloseHandle(hPipe) = False then
    begin
      ErrorMessageWindow;
      Exit;
    end;

    hPipe := INVALID_HANDLE_VALUE;
  end;

  Result := True;
end;

function TUniHandlerClient.CountCheckSum(const Data: array of Byte; const Index: integer; const Length: Word): Word;
var i: integer;
begin
  Result := 0;

  for i := 0 to Length - 1 do
    Result := Result + Data[Index + i];
end;

function TUniHandlerClient.WrapDataToPacket(var PacketData: array of Byte; const Buffer: string): Boolean;
var CheckSum: Word;
    i, iLength: integer;
begin
  Result := False;

  // check capability for PacketData length range
  iLength := Length(Buffer);
  if iLength > High(Word) then
    Exit;

  // set leading byte
  PacketData[0] := PACKET_LEADING_BYTE;

  // set data length
  PacketData[1] := Lo(iLength);
  PacketData[2] := Hi(iLength);

  // copy data of packet
  for i := 0 to iLength - 1 do
    PacketData[3 + i] := Byte(Buffer[i + 1]);

  // count check sum
  CheckSum := CountCheckSum(PacketData, 3, iLength);

  // set check sum
  PacketData[3 + iLength] := Lo(CheckSum);
  PacketData[3 + iLength + 1] := Hi(CheckSum);

  Result := True;
end;

function TUniHandlerClient.UnwrapDataFromPacket(var Buffer: string; const PacketData: array of Byte): Boolean;
var CheckSumFromPacket: Word;
    i, iLength: integer;
begin
  Result := False;
  Buffer := '';
  
  // check packet integrity
  if PacketData[0] <> PACKET_LEADING_BYTE then
    Exit;

  iLength := (PacketData[2] shl 8) + PacketData[1];
  CheckSumFromPacket := (PacketData[3 + iLength + 1] shl 8) + PacketData[3 + iLength];
  // compare packet check sum with calculated
  if CheckSumFromPacket <> CountCheckSum(PacketData, 3, iLength) then
    Exit;

  // copy data
  for i := 0 to iLength - 1 do
    Buffer := Buffer + Char(PacketData[3 + i]);

  Result := True;
end;

function TUniHandlerClient.Write(const Buffer: string): Boolean;
var lpNumberOfBytesWritten: DWORD;
    PacketData: array[0..PACKET_SIZE - 1] of Byte;
begin
  Result := False;

  if not WrapDataToPacket(PacketData, Buffer) then
    Exit;

  if hPipe = INVALID_HANDLE_VALUE then
    Exit;

  // write packet
  if not WriteFile(hPipe, PacketData, PACKET_PROTECTION_SIZE + Length(Buffer), lpNumberOfBytesWritten, nil) then
  begin
    ErrorMessageWindow;
    Exit;
  end;

  Result := True;
end;

function TUniHandlerClient.Read(var Buffer: string): Boolean;
var lpNumberOfBytesRead: DWORD;
    PacketData: array[0..PACKET_SIZE - 1] of Byte;
begin
  Result := False;
  Buffer := '';

  if hPipe = INVALID_HANDLE_VALUE then
    Exit;

  // read packet
  if not ReadFile(hPipe, PacketData, PACKET_SIZE, lpNumberOfBytesRead, nil) then
  begin
    ErrorMessageWindow;
    Exit;
  end;

  if not UnwrapDataFromPacket(Buffer, PacketData) then
  begin
    Buffer := TXT_NACK + TXT_NACK_DELIMITER + TXT_NACK_ERROR_1;
    Exit;
  end;

  Result := True;
end;

procedure TUniHandlerClient.WriteToLog(const Text: string);
begin
  if @pLogProcedure <> nil then
    pLogProcedure(PChar(Text));
end;

procedure TUniHandlerClient.ErrorMessageWindow;

  function MAKELANGID(p: Byte; s: Byte): Word;
  begin
    Result := (((Word(s)) shl 10) or Word(p))
  end;

var lpMsgBuf: HLOCAL;
begin
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_FROM_SYSTEM or
                FORMAT_MESSAGE_IGNORE_INSERTS,
                nil,
                GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                LPTSTR(@lpMsgBuf),
                0,
                nil);

  WriteToLog('Pipe Error: ' + Trim(LPCTSTR(lpMsgBuf)));
  // Free the buffer.
  LocalFree(lpMsgBuf);
end;

end.
