/*
	filename - CTFormVUH.h

	Purpose:
			 This file contains standard application window (form) and main control logic for Virtual Universal Handler (VUH).
*/
/*
History (dd.mm.yy):
21.01.2019 added new command "PROGRAMMING ENVIRONMENT"
10.12.2018 added new commands "SITES STATISTICS", "SITE STATISTICS"
18. 6.2012 added new command "SITE ZIF SOCKET EMPTY"
           checking of TXT_CMD_PROTOCOL_VERSION for required minimum value
21. 3.2012 added new parameters /JobOperatorID:, /JobID:, /ProjectPassword:, /UniqueID: for command "SITES LOAD PROJECT"
           added new results 7, 8, 9 for command "SITES LOAD PROJECT"
           change dialog of project file loading
 6. 2.2012 added new commands "SITES LOAD PROJECT", "APPLICATION CLOSE" and "SITE ACTUATOR STATUS"
           added option to start PG4UWMC remote control by command line parameters /autoconnectremotecontrol, /donotloadproject
17.10.2011 added new commands "SITES STOP" and "SITE MODULE CHECK N"
25. 7.2011 adding support for more as 8 site-s,
           adding new handler speed postition without pusher animation
 1.10.2010 start of development
*/
//---------------------------------------------------------------------------

#ifndef CTFormVUHH
#define CTFormVUHH
//---------------------------------------------------------------------------
#include <Classes.hpp>
#include <Controls.hpp>
#include <StdCtrls.hpp>
#include <Forms.hpp>
#include <ComCtrls.hpp>
#include <ExtCtrls.hpp>
#include <Graphics.hpp>

#include "CTUniHandlerClient.h"
#include "CTDedicatedDriver.h"
#include "CTSprite.h"
#include "CTReelSprite.h"
#include "CTPusherSprite.h"
#include "CTSocketSprite.h"
#include "CTStatistics.h"
#include "CTFormLoadProject.h"
#include <Dialogs.hpp>
//---------------------------------------------------------------------------

#define WM_USER_HANDLER_LOGIC_REALIZATION WM_USER + 1
#define WM_USER_HANDLER_LOGIC_REALIZATION_FINISHED WM_USER + 2

enum TSpriteElements { tSEReelIn = 0, tSEReelOutOK, tSEReelOutBad, tSEPusher, tSESocket };
enum TGUIControlState { tGCSBeforeInitialization = 0, tGCSAfterInitialization, tGCSBeforeStart, tGCSAfterStart };

class TFormVUH : public TForm
{
__published:	// IDE-managed Components
	TPanel *PanelStartUpServer;
	TLabel *LabelServerPath;
	TEdit *EditServerPath;
	TOpenDialog *OpenDialogServerPath;
	TButton *ButtonlServerPath;
	TCheckBox *CheckBoxAutoConnectRemoteControl;
	TCheckBox *CheckBoxDonotLoadProject;
	TButton *ButtonStartUp;
	TScrollBox *ScrollBoxPaintBox;
	TPaintBox *PaintBox;
	TImage *ImageReelPanelIn;
	TImage *ImageReelPanelOutOK;
	TImage *ImageReelPanelOutBad;
	TImage *ImagePusherEmpty;
	TImage *ImagePusherFull;
	TImage *ImageSocketDisabledEmpty;
	TImage *ImageSocketEnabledEmpty;
	TImage *ImageSocketEnabledFull;
	TImage *ImageSocketSideDisabledUp;
	TImage *ImageSocketSideEnabledUp;
	TImage *ImageSocketSideEnabledDown;
	TPanel *PanelControl;
	TComboBox *ComboBoxMode;
	TButton *ButtonInitialize;
	TButton *ButtonLoadProjectCheckModules;
	TPanel *PanelStart;
	TButton *ButtonStartStop;
	TPanel *PanelTotalStop;
	TButton *ButtonTotalStop;
	TButton *ButtonCloseServerApplication;
	TLabel *Label1;
	TTrackBar *TrackBarSpeed;
	TRichEdit *RichEditLogWindow;
	TLabel *LabelCmdLine;
	TEdit *EditCmdLine;
	void __fastcall FormShow(TObject *Sender);
	void __fastcall FormClose(TObject *Sender, TCloseAction &Action);
	void __fastcall FormCloseQuery(TObject *Sender, bool &CanClose);
	void __fastcall FormPaint(TObject *Sender);
	void __fastcall PaintBoxPaint(TObject *Sender);
	void __fastcall ButtonlServerPathClick(TObject *Sender);
	void __fastcall ButtonStartUpClick(TObject *Sender);
	void __fastcall ButtonInitializeClick(TObject *Sender);
	void __fastcall ButtonLoadProjectCheckModulesClick(TObject *Sender);
	void __fastcall ButtonStartStopClick(TObject *Sender);
	void __fastcall ButtonTotalStopClick(TObject *Sender);
	void __fastcall ButtonCloseServerApplicationClick(TObject *Sender);
	void __fastcall TrackBarSpeedChange(TObject *Sender);
	void __fastcall EditCmdLineKeyUp(TObject *Sender, WORD &Key,
					TShiftState Shift);

private:	// User declarations
	TUniHandlerClient *pUniHandlerClient; // universal handler
	TDedicatedDriver *pDedicatedDriver;
	TObjectList *pSpriteList;   			// sprite list of zif, input and output tube
	int iNumberSites;        					// number of sites of programmer
	int	iCurrentPosition;    					// position of handler
	Graphics::TBitmap *pBitmapMerged;	// merged bitmap background + pusher, reason: reduce blinking
	int	iStepSize;           					// define size of step for moving
	int	iSleepSize;          					// define size of sleep for moving
	bool bFullManualMode;     				// for future purpose
	TStrings *pKeyboardCommandsHistory;
	TGUIControlState FGUIControlState;

	TFormLoadProject *pFormLoadProject;

	void SetHandlerSpeed();
	void PaintSocket(const int Index);
	void EnableSocket(const int Index);
	void DisableSocket(const int Index);
	TSocketState GetSocketState(const int Index);
	bool MoveReel(const TSpriteElements SpriteElement);
	void LedBusy(const int Index, const bool EmitLight, const bool Redraw = true);
	void LedGood(const int Index, const bool EmitLight, const bool Redraw = true);
	void LedError(const int Index, const bool EmitLight, const bool Redraw = true);
	void WriteToLog(const AnsiString Text);
	void ResetStatistics();
	void IncrementReelStatistic(const TSpriteElements SpriteElement);
	void IncrementSocketStatistic(const int Index, const int ResultCode);
	MESSAGE void Finished(TMessage &Message);
	void ReconfigureHandler();
	void SetGUIControlState(const TGUIControlState AGUIControlState);

	bool SocketsEmpty();
	AnsiString ReportStatistics(AnsiString Text);
	// handler logic
	MESSAGE	void HandlerLogicRealization(TMessage &Message);

	// communication issues
	bool SendCommand(const AnsiString Command, AnsiString &Answer, const Boolean FromCmdLine = false);
	bool SendCommand(const AnsiString Command, const int Site, AnsiString &Answer);

	//// high level operation (Sequences)
	// insert (remove) device to (from) input (output) reel by pusher
	bool ReelOperationSequence(const TSpriteElements SpriteElement);
	// insert device to socket by pusher
	bool InsertSocketSequence(const int Index);
	// remove device from socket by pusher
	bool RemoveSocketSequence(const int Index);

	//// low level operation
	TPoint GetPosition(const int Index, const int ANumberSites);
	void PaintSprites();
	// move pusher from actual position to required NewPosition
	bool MovePusher(const int NewPosition, const int ANumberSites);
	// pull device by pusher
	bool PullDeviceByPusher();
	// drop device by pusher
	bool DropDeviceByPusher();
	// insert device to socket
	bool InsertDeviceToSocket(const int Index);
	// remove device from socket
	bool RemoveDeviceFromSocket(const int Index);
	int GetActuatorStatus(const int Index, const bool bPaint = true);

	void AddParameterToCommandLine(const AnsiString NewParameter, AnsiString &Parameters);
	void AddNoneBlankValueParameterToCommandLine(const AnsiString NewParameter, const AnsiString ValueOfNewParameter, AnsiString &Parameters);
	AnsiString ReportProgrammingEnvironment(AnsiString Text);

public:		// User declarations
	__property TGUIControlState GUIControlState = {read=FGUIControlState, write=FGUIControlState};

	__fastcall TFormVUH(TComponent* Owner);
	__fastcall ~TFormVUH();

BEGIN_MESSAGE_MAP
	MESSAGE_HANDLER(WM_USER_HANDLER_LOGIC_REALIZATION, TMessage, HandlerLogicRealization);
	MESSAGE_HANDLER(WM_USER_HANDLER_LOGIC_REALIZATION_FINISHED, TMessage, Finished);
END_MESSAGE_MAP(TForm);
};
//---------------------------------------------------------------------------
extern PACKAGE TFormVUH *FormVUH;
//---------------------------------------------------------------------------
#endif
