
#pragma hdrstop

#include <SysUtils.hpp>
#include "CTUniHandlerClient.h"

#define TXT_PIPE_NAME "\\\\.\\pipe\\PG4UWMCPipeName4RemoteControl"

#define PACKET_PROTECTION_SIZE 5 // Leading byte (1B), Data length (2B), Check sum (2B)
#define PACKET_SIZE 0xFFFF + PACKET_PROTECTION_SIZE
#define PACKET_LEADING_BYTE 0x7E

__fastcall TUniHandlerClient::TUniHandlerClient(const TLogProcedure pLogProcedureArg)
{
	hPipe = INVALID_HANDLE_VALUE;
	pLogProcedure = pLogProcedureArg;
}

__fastcall TUniHandlerClient::~TUniHandlerClient()
{
	Disconnect();  // disconnect pipe
}

bool TUniHandlerClient::Connect()
{
	if (!WaitNamedPipe(TXT_PIPE_NAME, NMPWAIT_USE_DEFAULT_WAIT))
	{
		ErrorMessageWindow();
		return false;
	}

	hPipe = CreateFile(TXT_PIPE_NAME, GENERIC_READ | GENERIC_WRITE, 0,
										 NULL,
										 OPEN_EXISTING,
										 0,
										 0);

	if (hPipe == INVALID_HANDLE_VALUE)
	{
		ErrorMessageWindow();
		return false;
	}

	return true;
}

bool TUniHandlerClient::Disconnect()
{
	if (hPipe != INVALID_HANDLE_VALUE)
	{
		if (CloseHandle(hPipe) == False)
		{
			ErrorMessageWindow();
			return false;
		}

		hPipe = INVALID_HANDLE_VALUE;
	}

	return true;
}

WORD TUniHandlerClient::CountCheckSum(const byte Data[], const int Index, const WORD Length)
{
	WORD Result = 0;

	for (int i = 0; i < Length; i++)
		Result += Data[Index + i];
	return Result;
}

bool TUniHandlerClient::WrapDataToPacket(byte PacketData[], const AnsiString Buffer)
{
	// check capability for PacketData length range
	int iLength = Buffer.Length();
	if (iLength > 0xFFFF)
		return false;

	// set leading byte
	PacketData[0] = PACKET_LEADING_BYTE;

	// set data length
	PacketData[1] = LOBYTE(iLength);
	PacketData[2] = HIBYTE(iLength);

	// copy data of packet
	for (int i = 0; i < iLength; i++)
		PacketData[3 + i] = (Byte)Buffer[i + 1];

	// count check sum
	WORD CheckSum = CountCheckSum(PacketData, 3, iLength);

	// set check sum
	PacketData[3 + iLength] = LOBYTE(CheckSum);
	PacketData[3 + iLength + 1] = HIBYTE(CheckSum);

	return true;
}

bool TUniHandlerClient::UnwrapDataFromPacket(AnsiString &Buffer, const byte PacketData[])
{
	Buffer = "";
	
	// check packet integrity
	if (PacketData[0] != PACKET_LEADING_BYTE)
		return false;

	int iLength = (PacketData[2] << 8) + PacketData[1];
	WORD CheckSumFromPacket = (PacketData[3 + iLength + 1] << 8) + PacketData[3 + iLength];
	// compare packet check sum with calculated
	if (CheckSumFromPacket != CountCheckSum(PacketData, 3, iLength))
		return false;

	// copy data
	for (int i = 0; i < iLength; i++)
		Buffer = Buffer + (Char)PacketData[3 + i];

  return true;
}

bool TUniHandlerClient::Write(const AnsiString Buffer)
{
	DWORD lpNumberOfBytesWritten;
	byte PacketData[PACKET_SIZE];

	if (!WrapDataToPacket(PacketData, Buffer))
		return false;

	if (hPipe == INVALID_HANDLE_VALUE)
		return false;

	// write packet		
	if (!WriteFile(hPipe, PacketData, PACKET_PROTECTION_SIZE + Buffer.Length(), &lpNumberOfBytesWritten, NULL))
	{
		ErrorMessageWindow();
		return false;
	}
	
	return true;
}

bool TUniHandlerClient::Read(AnsiString &Buffer)
{
	DWORD lpNumberOfBytesRead;
	byte PacketData[PACKET_SIZE];
	Buffer = "";

	if (hPipe == INVALID_HANDLE_VALUE)
		return false;

	// read packet
	if (!ReadFile(hPipe, PacketData, PACKET_SIZE, &lpNumberOfBytesRead, NULL))
	{
		ErrorMessageWindow();
		return false;
	}

	if (!UnwrapDataFromPacket(Buffer, PacketData))
	{
		Buffer = TXT_NACK + (AnsiString)TXT_NACK_DELIMITER + TXT_NACK_ERROR_1;
		return false;
	}
	
	return true;
}

void TUniHandlerClient::WriteToLog(const AnsiString Text)
{
	if (pLogProcedure != NULL)
		pLogProcedure(Text.c_str());
}

void TUniHandlerClient::ErrorMessageWindow()
{
	HLOCAL lpMsgBuf;

	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
								FORMAT_MESSAGE_IGNORE_INSERTS,
								NULL,
								GetLastError(),
								MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
								(LPTSTR)&lpMsgBuf,
								0,
								NULL);
	WriteToLog("Pipe Error: " + Trim((AnsiString)LPCTSTR(lpMsgBuf)));
	// Free the buffer.
	LocalFree(lpMsgBuf);
}


#pragma package(smart_init)
