﻿/*
	filename - FormVUH.cs

	Purpose:
			 This file contains standard application window (form) and main control logic for Virtual Universal Handler (VUH).
*/
/*
History (dd.mm.yy):
21.01.2019 added new command "PROGRAMMING ENVIRONMENT" 
10.12.2018 added new commands "SITES STATISTICS", "SITE STATISTICS"
18. 6.2012 added new command "SITE ZIF SOCKET EMPTY"
           checking of TXT_CMD_PROTOCOL_VERSION for required minimum value 
21. 3.2012 added new parameters /JobOperatorID:, /JobID:, /ProjectPassword:, /UniqueID: for command "SITES LOAD PROJECT"
           added new results 7, 8, 9 for command "SITES LOAD PROJECT"
           change dialog of project file loading 
 6. 2.2012 added new commands "SITES LOAD PROJECT", "APPLICATION CLOSE" and "SITE ACTUATOR STATUS"
           added option to start PG4UWMC remote control by command line parameters /autoconnectremotecontrol, /donotloadproject
17.10.2011 added new commands "SITES STOP" and "SITE MODULE CHECK N"
25. 7.2011 adding support for more as 8 site-s,
           adding new handler speed postition without pusher animation 
 1.10.2010 start of development					 
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;

namespace VUH
{
    public partial class FormVUH : Form
    {
        public delegate void EventHandlerLogicRealization();
        public event EventHandlerLogicRealization OnEventHandlerLogicRealization;
        public delegate void EventHandlerLogicRealizationFinished(bool bCorrectFinish);
        public event EventHandlerLogicRealizationFinished OnEventHandlerLogicRealizationFinished;

        enum EnumSpriteElements { tSEReelIn = 0, tSEReelOutOK, tSEReelOutBad, tSEPusher, tSESocket };
        enum EnumGUIControlState { tGCSBeforeInitialization = 0, tGCSAfterInitialization, tGCSBeforeStart, tGCSAfterStart };        

        const string TXT_TRUE = "1";
        const string TXT_FALSE = "0";

        const string TXT_CMD_PROTOCOL_VERSION           = "PROTOCOL VERSION";
        const string TXT_CMD_PG4UWMC_VERSION            = "PG4UWMC VERSION";
        const string TXT_CMD_MAX_INDEX_RANGE            = "SITES COUNT";
        const string TXT_CMD_APPLICATION_CLOSE          = "APPLICATION CLOSE";
        const string TXT_CMD_PROGRAMMING_ENVIRONMENT    = "PROGRAMMING ENVIRONMENT";

        const string TXT_CMD_SITES_STOP                 = "SITES STOP";
        const string TXT_CMD_SITES_LOAD_PROJECT         = "SITES LOAD PROJECT";

        const string TXT_CMD_IDX_SITE_PRESENT           = "SITE PRESENT";  // state of site = APP_PRESENT and PGM_STATUS
        const string TXT_CMD_IDX_SITE_BUSY              = "SITE BUSY";
        const string TXT_CMD_IDX_SITE_RESULT            = "SITE RESULT";
        const string TXT_CMD_IDX_SITE_START             = "SITE START";
        const string TXT_CMD_IDX_SITE_SERIALNUMBER      = "SITE SERIALNUMBER";
        const string TXT_CMD_IDX_SITE_MODULE_CHECK      = "SITE MODULE CHECK";
        const string TXT_CMD_IDX_SITE_ACTUATOR_STATUS   = "SITE ACTUATOR STATUS";
        const string TXT_CMD_IDX_SITE_ZIF_SOCKET_EMPTY  = "SITE ZIF SOCKET EMPTY";
        const string TXT_CMD_IDX_SITES_STATISTICS      	= "SITES STATISTICS";
        const string TXT_CMD_IDX_SITE_STATISTICS        = "SITE STATISTICS";

        const string TXT_CMD_UNKNOWN                    = "UNKNOWN COMMAND";

        const string TXT_HANDLER_RUNNING = "Handler is still running!\nPlease wait a moment while handler correctly stop.";

        const string TXT_START = "Start";
        const string TXT_STOP = "Stop";

        const string TXT_MESSAGEBOX_TITLE = "Virtual Universal Handler";

        const int DEFAULT_NUMBER_SITES = 8;

        const string TXT_MINIMUM_REQUIRED_PROTOCOL_VERSION = "1.11";

        bool bSemaphore = true;      // indicate running activity of procedure, protect procedure from multiple calling
        int constIndexInKeyboardCommandsHistory = 0;

        private UniHandlerClient uniHandlerClient = null;
        private DedicatedDriver dedicatedDriver;
        private List<Sprite> spriteList;    // sprite list of zif, input and output tube
        private int iNumberSites;           // number of sites of programmer
        private int iCurrentPosition;       // position of handler
        private Bitmap bitmapMerged;        // merged bitmap background + pusher, reason: reduce blinking
        private int iStepSize;              // define size of step for moving
        private int iSleepSize;             // define size of sleep for moving
        private bool bFullManualMode;       // for future purpose
        private List<string> keyboardCommandsHistory;
        EnumGUIControlState fGUIControlState;
        EnumGUIControlState GUIControlState
        {
            get { return fGUIControlState; }
            set { fGUIControlState = value; }
        }

        private FormLoadProject formLoadProject;
        
        public FormVUH()
        {
            InitializeComponent();

            spriteList = new List<Sprite>();
            dedicatedDriver = new DedicatedDriver(WriteToLog);

            iNumberSites = DEFAULT_NUMBER_SITES;

            iCurrentPosition = 0;
            iStepSize = 1;
            iSleepSize = 0;

            bitmapMerged = new Bitmap(pictureBox.Size.Width, pictureBox.Size.Height);

            keyboardCommandsHistory = new List<string>();

            fGUIControlState = EnumGUIControlState.tGCSBeforeInitialization;

            bFullManualMode = false;

            formLoadProject = new FormLoadProject();

            uniHandlerClient = new UniHandlerClient(WriteToLog);
            OnEventHandlerLogicRealization += HandlerLogicRealization;
            OnEventHandlerLogicRealizationFinished += Finished;            
        }
        private void FormVUH_Shown(object sender, EventArgs e)
        {
            comboBoxMode.SelectedIndex = 0;

            ReconfigureHandler();

            SetGUIControlState(EnumGUIControlState.tGCSBeforeInitialization);

            SetHandlerSpeed();
        }
        private void FormVUH_FormClosing(object sender, FormClosingEventArgs e)
        {
            e.Cancel = !((GUIControlState == EnumGUIControlState.tGCSBeforeInitialization) || (GUIControlState == EnumGUIControlState.tGCSAfterInitialization));
            if (e.Cancel)
                MessageBox.Show(TXT_HANDLER_RUNNING, TXT_MESSAGEBOX_TITLE, MessageBoxButtons.OK);
        }
        private void FormVUH_Paint(object sender, PaintEventArgs e)
        {
            if (spriteList.Count <= 0)
                return;

            // draw background            
            bitmapMerged = new Bitmap(pictureBox.Size.Width, pictureBox.Size.Height);
            Graphics graphics = Graphics.FromImage(bitmapMerged);
            graphics.FillRectangle(new SolidBrush(SystemColors.ButtonFace), 0, 0, bitmapMerged.Size.Width, bitmapMerged.Size.Height);            
            // draw sprites
            for (int i = 0; i < spriteList.Count; i++)
                (spriteList[i] as Sprite).Paint(ref graphics);
            pictureBox.Image = bitmapMerged;
            graphics.Dispose();
        }
        private void buttonServerPath_Click(object sender, EventArgs e)
        {
            const string TXT_PG4UWMC = "PG4UWMC.exe";

            openFileDialogServerPath.Filter = "Program " + TXT_PG4UWMC + "|" + TXT_PG4UWMC;
            openFileDialogServerPath.CheckPathExists = true;
            openFileDialogServerPath.CheckFileExists = true;
            if (openFileDialogServerPath.ShowDialog() == DialogResult.OK)
                textBoxServerPath.Text = openFileDialogServerPath.FileName;
        }
        private void AddParameterToCommandLine(string newParameter, ref string parameters)
        {
            if (parameters != "")
                parameters = parameters + " ";
            parameters = parameters + newParameter;
        }
        private void buttonStartUp_Click(object sender, EventArgs e)
        {
            const string TXT_CMDLINE_AUTOCONNECTREMOTECONTROL = "/autoconnectremotecontrol";
            const string TXT_CMDLINE_DONOTLOADPROJECT = "/donotloadproject";

            string fileNameWithPath, parameters, sTextTemp;

            WriteToLog("Starting remote control server");

            fileNameWithPath = textBoxServerPath.Text;
            if (!File.Exists(fileNameWithPath))
            {
                sTextTemp = "Server: '" + fileNameWithPath + "' does not exist.\n" +
                            "Please set correct server with path.";
                WriteToLog(sTextTemp);
                MessageBox.Show(sTextTemp, TXT_MESSAGEBOX_TITLE, MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return;
            }

            parameters = "";
            if (checkBoxAutoConnectRemoteControl.Checked == true)
                AddParameterToCommandLine(TXT_CMDLINE_AUTOCONNECTREMOTECONTROL, ref parameters);
            if (checkBoxDonotLoadProject.Checked == true)
                AddParameterToCommandLine(TXT_CMDLINE_DONOTLOADPROJECT, ref parameters);

            try
            {
                Process.Start(fileNameWithPath, parameters);
            }
            catch (Exception exception)
            {
                sTextTemp = "Application: '" + fileNameWithPath + "' can not be started.\n" + exception.Message;
                WriteToLog(sTextTemp);
                MessageBox.Show(sTextTemp, TXT_MESSAGEBOX_TITLE, MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return;
            }
            WriteToLog("Application: '" + fileNameWithPath + " " + parameters + "' run successfully.");
        }
        private void buttonInitialize_Click(object sender, EventArgs e)
        {
            if (bFullManualMode == false)
            {
                // stop of universal handler client
                if (uniHandlerClient != null)
                {
                    uniHandlerClient.Disconnect();
                    Thread.Sleep(100);  // wait for server
                }

                if (uniHandlerClient.Connect() == false)
                    return;
            }

            string textTemp = "";
            // mandatory as first instruction in code
            if (!SendCommand(TXT_CMD_PROTOCOL_VERSION, ref textTemp))
                return;
            if (String.Compare(TXT_MINIMUM_REQUIRED_PROTOCOL_VERSION, textTemp, true) > 0)
            {
                WriteToLog("Minimum required protocol version (" + TXT_MINIMUM_REQUIRED_PROTOCOL_VERSION +
                           ") is greater than the current running PG4UWMC protocol version (" + textTemp + ").");
                WriteToLog("Please close and update PG4UWMC.");
                return;
            }

            if (!SendCommand(TXT_CMD_PG4UWMC_VERSION, ref textTemp))
                return;

            if (!SendCommand(TXT_CMD_MAX_INDEX_RANGE, ref textTemp))
                return;           

            try
            {
                iNumberSites = Convert.ToInt32(textTemp);
            }
            catch (FormatException)
            {
                iNumberSites = DEFAULT_NUMBER_SITES;
            }

            ReconfigureHandler();
            for (int i = 0; i < iNumberSites; i++)
            {
                // get actuator status
                if (GetActuatorStatus(i, false) == -1)
                    return;

                // get site presence
                if (!SendCommand(TXT_CMD_IDX_SITE_PRESENT, i, ref textTemp))
                    return;
                if (textTemp != TXT_TRUE)
                    DisableSocket(i);
                else
                    // site is present
                    EnableSocket(i);

                Application.DoEvents();
            }
            this.Refresh();

            // get sites serial number
            for (int i = 0; i < iNumberSites; i++)
            {
                if (!SendCommand(TXT_CMD_IDX_SITE_SERIALNUMBER, i, ref textTemp))
                    return;
                (spriteList[i] as SocketSprite).SerialNumber = textTemp;
            }

            // check modules in sites
            for (int i = 0; i < iNumberSites; i++)
            {
                if (!SendCommand(TXT_CMD_IDX_SITE_MODULE_CHECK, i, ref textTemp))
                    return;
                Application.DoEvents();                
            }

            // check if ZIF socket is empty
            for (int i = 0; i < iNumberSites; i++)
            {
                if (!SendCommand(TXT_CMD_IDX_SITE_ZIF_SOCKET_EMPTY, i, ref textTemp))
                    return;
                Application.DoEvents();
            }

            this.Refresh();

            if (dedicatedDriver.HandlerActive())
                SetGUIControlState(EnumGUIControlState.tGCSAfterInitialization);
        }
        private void AddNoneBlankValueParameterToCommandLine(string newParameter, string valueOfNewParameter, ref string parameters)
        {
	        if (valueOfNewParameter != "")
	        {
		        if (parameters != "")
			        parameters = parameters + " ";
		        parameters = parameters + newParameter + valueOfNewParameter;
	        }
        }
        private string ReportProgrammingEnvironment(string text)
        {
	        int iIndexInText;
	        string sItem;

	        string result = "Report of programming environment:\n";

	        do
	        {
              iIndexInText = text.IndexOf('|');
	          if (iIndexInText > 0)
	          {
		        sItem = text.Substring(0, iIndexInText);
                text = text.Remove(0, iIndexInText + 1);
		        result = result + sItem + '\n';
	          }
	        } while (text != "");

	        return result;
        }
        private void buttonLoadProjectCheckModules_Click(object sender, EventArgs e)
        {
            string projectFileNameWithPath = "", sAnswer = "", sTextTemp;
            int i, iAnswer;
            bool checkModules = true;
            string optionalParamaters, jobOperatorID = "", jobID = "", projectPassword = "", uniqueID = "";

            if (formLoadProject.ShowDialog() == DialogResult.OK)
            {
                formLoadProject.GetParameter(ref checkModules, ref projectFileNameWithPath, ref jobOperatorID, ref jobID, ref projectPassword, ref uniqueID);
                WriteToLog("Load project: " + projectFileNameWithPath);

                optionalParamaters = "";
                AddNoneBlankValueParameterToCommandLine("/JobOperatorID:", jobOperatorID, ref optionalParamaters);
                AddNoneBlankValueParameterToCommandLine("/JobID:", jobID, ref optionalParamaters);
                AddNoneBlankValueParameterToCommandLine("/ProjectPassword:", projectPassword, ref optionalParamaters);
                AddNoneBlankValueParameterToCommandLine("/UniqueID:", uniqueID, ref optionalParamaters);
                if (optionalParamaters != "")
                    optionalParamaters = " " + optionalParamaters;
                if (!SendCommand(TXT_CMD_SITES_LOAD_PROJECT + ' ' + projectFileNameWithPath + optionalParamaters, ref sAnswer))
                    return;

                // convert answer from string to integer
                try
                {
                    iAnswer = Convert.ToInt32(sAnswer);
                }
                catch (FormatException)
                {
                    iAnswer = 0;
                }

                // process answer of command
                switch (iAnswer)
                {
                    case 0: sTextTemp = "Error: Unknown."; break;
                    case 1: sTextTemp = "Project file load successful."; break;
                    case 2: sTextTemp = "Error: Check 'Use Site #1 project for all Sites' in 'PG4UWMC Settings' dialog."; break;
                    case 3: sTextTemp = "Error: Some site(s) is busy, stop programming operation before loading."; break;
                    case 4: sTextTemp = "Error: Master site was not correctly started.\n" +
                                  "       Try to do following steps in PG4UWMC:\n" +
                                  "         1. disconnect Sites\n" +
                                  "         2. connect Sites\n" +
                                  "         3. load project."; break;
                    case 5: sTextTemp = "Error: Failure during loading process."; break;
                    case 6: sTextTemp = "Error: Bad project file password."; break;
                    case 7: sTextTemp = "Error: Project load forbidden"; break;
                    case 8: sTextTemp = "Project file load successful, but forced default \"Program\" operation."; break;
                    case 9: sTextTemp = "Error: Bad project file unique ID."; break;
                    default: sTextTemp = "Error: Unknown."; break;
                }
                WriteToLog(sTextTemp);
                if ((iAnswer != 1) && (iAnswer != 8))
                {
                    MessageBox.Show(sTextTemp, TXT_MESSAGEBOX_TITLE, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    return;
                }
                else
                {
                    // read and report programming environment
                    if (SendCommand(TXT_CMD_PROGRAMMING_ENVIRONMENT, ref sAnswer))
                        WriteToLog(ReportProgrammingEnvironment(sAnswer));
                }


                if (checkModules)
                {
                    // check modules in sites
                    sTextTemp = "";
                    for (i = 0; i < iNumberSites; i++)
                    {
                        if (!SendCommand(TXT_CMD_IDX_SITE_MODULE_CHECK, i, ref sAnswer))
                            return;
                        if (sAnswer != TXT_TRUE)
                            if (sTextTemp != "")
                                sTextTemp = sTextTemp + ", #" + (i + 1).ToString();
                            else
                                sTextTemp = sTextTemp + "#" + (i + 1).ToString();
                        Application.DoEvents();
                    }
                    if (sTextTemp != "")
                    {
                        sTextTemp = "AP module(s) errors summary: " + sTextTemp + ".";
                        WriteToLog(sTextTemp);
                        MessageBox.Show(sTextTemp, TXT_MESSAGEBOX_TITLE, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    }
                    else
                        WriteToLog("AP module(s) check successful.");
                }
            }
        }
        private void buttonStartStop_Click(object sender, EventArgs e)
        {
            if (GUIControlState == EnumGUIControlState.tGCSAfterInitialization)
            {
                SetGUIControlState(EnumGUIControlState.tGCSBeforeStart);

                // run Handler Logic Realization
                if (OnEventHandlerLogicRealization != null)
                    OnEventHandlerLogicRealization();
            }
            else
                SetGUIControlState(EnumGUIControlState.tGCSAfterStart);
        }
        private void buttonTotalStop_Click(object sender, EventArgs e)
        {
            string sAnswer = "";

            WriteToLog("Total-Stop pressed.");
            if (!SendCommand(TXT_CMD_SITES_STOP, ref sAnswer))
                return;
        }
        private void buttonCloseServerApplication_Click(object sender, EventArgs e)
        {
            string sAnswer, sTextTemp;
            int iAnswer;

            WriteToLog("Close Application button pressed.");
            sAnswer = "";
            if (!SendCommand(TXT_CMD_APPLICATION_CLOSE, ref sAnswer))
                return;

            // convert answer from string to integer
            try
            {
                iAnswer = Convert.ToInt32(sAnswer);
            }
            catch (FormatException)
            {
                iAnswer = 0;
            }

            // process answer of command
            switch (iAnswer)
            {
                case 0: sTextTemp = "Error: Unknown."; break;
                case 1: sTextTemp = "Close application successful."; break;
                case 2: sTextTemp = "Error: Some site(s) is busy."; break;
                case 3: sTextTemp = "Error: Impossible to disconnect site(s)."; break;
                default: sTextTemp = "Error: Unknown."; break;
            }
            WriteToLog(sTextTemp);
            if (iAnswer != 1)
                MessageBox.Show(sTextTemp, TXT_MESSAGEBOX_TITLE, MessageBoxButtons.OK, MessageBoxIcon.Warning);
        }
        private void trackBarSpeed_ValueChanged(object sender, EventArgs e)
        {
            SetHandlerSpeed();
        }
        private void textBoxCmdLine_KeyUp(object sender, KeyEventArgs e)
        {            
            string sAnswer = "";

            // set last command when UP ARROW key is press
            if (e.KeyCode == Keys.Up)
            {
                // select previous command from history
                if (keyboardCommandsHistory.Count > 0)
                    if (constIndexInKeyboardCommandsHistory < keyboardCommandsHistory.Count)
                    {
                        constIndexInKeyboardCommandsHistory++;
                        textBoxCmdLine.Text = keyboardCommandsHistory[constIndexInKeyboardCommandsHistory - 1];
                    }
                textBoxCmdLine.Select(textBoxCmdLine.Text.Length, textBoxCmdLine.Text.Length);
            }
            // set last command when DOWN ARROW key is press
            else if (e.KeyCode == Keys.Down)
            {
                // select following command from history
                if (keyboardCommandsHistory.Count > 0)
                    if (constIndexInKeyboardCommandsHistory > 1)
                    {
                        constIndexInKeyboardCommandsHistory--;
                        textBoxCmdLine.Text = keyboardCommandsHistory[constIndexInKeyboardCommandsHistory - 1];
                    }
                textBoxCmdLine.Select(textBoxCmdLine.Text.Length, textBoxCmdLine.Text.Length);
            }
            else if (e.KeyCode == Keys.Return)
            {
                if (!SendCommand(textBoxCmdLine.Text, ref sAnswer, true))
                {}

                // insert command to history
                if (keyboardCommandsHistory.Count == 0)
                    keyboardCommandsHistory.Add(textBoxCmdLine.Text);
                else if (keyboardCommandsHistory[0] != textBoxCmdLine.Text)
                    keyboardCommandsHistory.Insert(0, textBoxCmdLine.Text);
                constIndexInKeyboardCommandsHistory = 0;

                // clear cmd line
                textBoxCmdLine.Text = "";
            }
        }

        private void SetHandlerSpeed()
        {
            switch (trackBarSpeed.Value)
            {
                case 1: iStepSize = 1; iSleepSize = 2; break;
                case 2: iStepSize = 1; iSleepSize = 1; break;
                case 3: iStepSize = 1; iSleepSize = 0; break;
                case 4: iStepSize = 2; iSleepSize = 0; break;
                case 5: iStepSize = 4; iSleepSize = 0; break;
                case 6: iStepSize = 5; iSleepSize = 0; break;
                default: iStepSize = 1; iSleepSize = 0; break;
            }
            (spriteList[iNumberSites + (int)EnumSpriteElements.tSEReelIn] as ReelSprite).SetDeltaMoving(iStepSize);
            (spriteList[iNumberSites + (int)EnumSpriteElements.tSEReelOutOK] as ReelSprite).SetDeltaMoving(iStepSize);
            (spriteList[iNumberSites + (int)EnumSpriteElements.tSEReelOutBad] as ReelSprite).SetDeltaMoving(iStepSize);
        }
        private void PaintSocket(int index)
        {
            Graphics graphics = Graphics.FromImage(bitmapMerged);
            (spriteList[index] as SocketSprite).Paint(ref graphics);
            (spriteList[iNumberSites + (int)EnumSpriteElements.tSEPusher] as PusherSprite).Paint(ref graphics);
            pictureBox.Image = bitmapMerged;
            graphics.Dispose();
        }
        private void EnableSocket(int index)
        {
            if ((spriteList[index] as SocketSprite).SocketState == EnumSocketState.tSDisabledEmpty)
            {
                (spriteList[index] as SocketSprite).SocketState = EnumSocketState.tSEnabledEmpty;
                PaintSocket(index);
            }
        }
        private void DisableSocket(int index)
        {
            if ((spriteList[index] as SocketSprite).SocketState == EnumSocketState.tSDisabledEmpty)
                return;

            (spriteList[index] as SocketSprite).SocketState = EnumSocketState.tSDisabledEmpty;
            LedBusy(index, false, false);
            LedGood(index, false, false);
            LedError(index, false, false);
            PaintSocket(index);
        }
        private EnumSocketState GetSocketState(int index)
        {
            return (spriteList[index] as SocketSprite).SocketState;
        }
        private bool MoveReel(EnumSpriteElements spriteElement)
        {          
            switch (spriteElement)
            {
                case EnumSpriteElements.tSEReelIn:
                case EnumSpriteElements.tSEReelOutOK:
                case EnumSpriteElements.tSEReelOutBad: break;
                default: return false;
            }

            (spriteList[iNumberSites + (int)spriteElement] as ReelSprite).MoveFrame();
                return true;
        }
        private void LedBusy(int index, bool emitLight, bool redraw = true)
        {
            (spriteList[index] as SocketSprite).LedBusy = emitLight;
            if (redraw)
                PaintSocket(index);
        }
        private void LedGood(int index, bool emitLight, bool redraw = true)
        {
            (spriteList[index] as SocketSprite).LedGood = emitLight;
            if (redraw)
                PaintSocket(index);
        }
        private void LedError(int index, bool emitLight, bool redraw = true)
        {
            (spriteList[index] as SocketSprite).LedError = emitLight;
            if (redraw)
                PaintSocket(index);
        }
        private void WriteToLog(string text)
        {
            richTextBoxLogWindow.AppendText(text + "\n");
            richTextBoxLogWindow.ScrollToCaret();
        }
        private void ResetStatistics()
        {
            (spriteList[iNumberSites + (int)EnumSpriteElements.tSEReelIn] as ReelSprite).Statistics = 0;
            (spriteList[iNumberSites + (int)EnumSpriteElements.tSEReelOutOK] as ReelSprite).Statistics = 0;
            (spriteList[iNumberSites + (int)EnumSpriteElements.tSEReelOutBad] as ReelSprite).Statistics = 0;

            for (int i = 0; i < iNumberSites; i++)
                (spriteList[i] as SocketSprite).Statistics.Reset();
        }
        private void IncrementReelStatistic(EnumSpriteElements spriteElement)
        {
            switch (spriteElement)
            {
                case EnumSpriteElements.tSEReelIn:
                case EnumSpriteElements.tSEReelOutOK:
                case EnumSpriteElements.tSEReelOutBad: break;
                default: return;
            }

            (spriteList[iNumberSites + (int)spriteElement] as ReelSprite).Statistics++;
        }
        private void IncrementSocketStatistic(int index, int resultCode)
        {
            (spriteList[index] as SocketSprite).Statistics.IncrementByResult(resultCode);
        }
        private void Finished(bool bCorrectFinish)
        {
            if (bCorrectFinish)
              SetGUIControlState(EnumGUIControlState.tGCSAfterInitialization);
            else
              SetGUIControlState(EnumGUIControlState.tGCSBeforeInitialization);
        }
        private void ReconfigureHandler()
        {
            const int x0 = 10;
            const int y0 = 10;

            spriteList.Clear();

            float sizeHorizontalElement = ((Bitmap)(pictureBoxSocketDisabledEmpty.Image)).Width * 1.3f;

            pictureBox.Width = x0 + (int)(Math.Max(iNumberSites, 8) * sizeHorizontalElement);
            MaximumSize = new System.Drawing.Size(Width - panelPaintBox.Width + pictureBox.Width + (2 * System.Windows.Forms.SystemInformation.Border3DSize.Width), System.Windows.Forms.SystemInformation.PrimaryMonitorSize.Height);
            
            // adding sites to list
            for (int i = 0; i < iNumberSites; i++)
                spriteList.Add(new SocketSprite(x0 + (int)(i * sizeHorizontalElement), 20 * y0, i + 1, 
                               new Bitmap(pictureBoxSocketDisabledEmpty.Image), new Bitmap(pictureBoxSocketEnabledEmpty.Image), new Bitmap(pictureBoxSocketEnabledFull.Image),
                               new Bitmap(pictureBoxSocketSideDisabledUp.Image), new Bitmap(pictureBoxSocketSideEnabledUp.Image), new Bitmap(pictureBoxSocketSideEnabledDown.Image)));

            sizeHorizontalElement = ((Bitmap)(pictureBoxReelPanelIn.Image)).Width * 3.0f / 4.0f * 1.2f;
            // adding other elements to list
            // ReelIn
            spriteList.Add(new ReelSprite(x0 + (int)((float)EnumSpriteElements.tSEReelIn * sizeHorizontalElement), 1 * y0, new Bitmap(pictureBoxReelPanelIn.Image)));
            // ReelOutOK
            spriteList.Add(new ReelSprite(x0 + (int)((float)EnumSpriteElements.tSEReelOutOK * sizeHorizontalElement), 1 * y0, new Bitmap(pictureBoxReelPanelOutOK.Image)));
            // ReelOutBad
            spriteList.Add(new ReelSprite(x0 + (int)((float)EnumSpriteElements.tSEReelOutBad * sizeHorizontalElement), 1 * y0, new Bitmap(pictureBoxReelPanelOutBad.Image)));
            // Pusher
            spriteList.Add(new PusherSprite(x0 + (int)((float)EnumSpriteElements.tSEReelIn * sizeHorizontalElement), 1 * y0, new Bitmap(pictureBoxPusherEmpty.Image), new Bitmap(pictureBoxPusherFull.Image)));

            ResetStatistics();

            iCurrentPosition = iNumberSites + (int)EnumSpriteElements.tSEReelIn;
        }
        private void SetGUIControlState(EnumGUIControlState aGUIControlState)
        {
            switch (aGUIControlState)
            {
                case EnumGUIControlState.tGCSBeforeInitialization: 
                          comboBoxMode.Enabled = true;

                          buttonInitialize.Enabled = true;

                          buttonStartStop.Text = TXT_START;
                          buttonStartStop.Enabled = false;
                          panelStart.BackColor = Color.Green;

                          buttonLoadProjectCheckModules.Enabled = false;
                          buttonTotalStop.Enabled = false;
                          buttonCloseServerApplication.Enabled = false;
                          panelTotalStop.BackColor = Color.Maroon;

                          labelCmdLine.Enabled = false;
                          textBoxCmdLine.Enabled = false;
                        break;
                case EnumGUIControlState.tGCSAfterInitialization:
                          comboBoxMode.Enabled = true;

                          buttonInitialize.Enabled = true;

                          buttonStartStop.Text = TXT_START;
                          buttonStartStop.Enabled = true;
                          panelStart.BackColor = Color.Lime;

                          buttonLoadProjectCheckModules.Enabled = true;
                          buttonTotalStop.Enabled = true;
                          buttonCloseServerApplication.Enabled = true;
                          panelTotalStop.BackColor = Color.Red;


                          labelCmdLine.Enabled = true;
                          textBoxCmdLine.Enabled = true;
                        break;
                case EnumGUIControlState.tGCSBeforeStart:
                          comboBoxMode.Enabled = false;

                          buttonInitialize.Enabled = false;

                          buttonStartStop.Text = TXT_STOP;
                          buttonStartStop.Enabled = true;
                          panelStart.BackColor = Color.Lime;

                          buttonLoadProjectCheckModules.Enabled = false;
                          buttonTotalStop.Enabled = true;
                          buttonCloseServerApplication.Enabled = false;
                          panelTotalStop.BackColor = Color.Red;

                          labelCmdLine.Enabled = false;
                          textBoxCmdLine.Enabled = false;
                        break;
                case EnumGUIControlState.tGCSAfterStart:
                          comboBoxMode.Enabled = false;
                          
                          buttonInitialize.Enabled = false;

                          buttonStartStop.Text = TXT_STOP;
                          buttonStartStop.Enabled = false;
                          panelStart.BackColor = Color.Green;

                          buttonLoadProjectCheckModules.Enabled = false;
                          buttonTotalStop.Enabled = true;
                          buttonCloseServerApplication.Enabled = false;
                          panelTotalStop.BackColor = Color.Red;

                          labelCmdLine.Enabled = false;
                          textBoxCmdLine.Enabled = false;
                        break;
                default: return;
            }
            fGUIControlState = aGUIControlState;
        }

        private bool SocketsEmpty()
        {
            for (int i = 0; i < iNumberSites; i++)
                if (GetSocketState(i) == EnumSocketState.tSEnabledFull)
                    return false;

            return true;
        }

        string ReportStatistics(string text)
        {
	        int iIndexInText, iStatisticValue;
	        string sStatisticItem;
	        int iStatisticSuccess, iStatisticAdapterTestFailure, iStatisticInsertionTestFailure, iStatisticIDCheckFailure,
		        iStatisticAccessCanceledByUserFailure, iStatisticOtherFailure, iStatisticOperationalFail;

	        iStatisticSuccess = 0;
	        iStatisticAdapterTestFailure = 0;
	        iStatisticInsertionTestFailure = 0;
	        iStatisticIDCheckFailure = 0;
	        iStatisticAccessCanceledByUserFailure = 0;
	        iStatisticOtherFailure = 0;
	        iStatisticOperationalFail = 0;

	        do
	        {
	            iIndexInText = text.IndexOf('|');
	            if (iIndexInText > -1)
	            {
		            sStatisticItem = text.Substring(0, iIndexInText);
                    text = text.Remove(0, iIndexInText + 1);

		            iIndexInText = sStatisticItem.IndexOf(':');
		            if (iIndexInText > -1)
		            {
                        iStatisticValue = Convert.ToInt32(sStatisticItem.Substring(iIndexInText + 1, sStatisticItem.Length - (iIndexInText + 1)));
                      switch (Convert.ToInt32(sStatisticItem.Substring(0, iIndexInText)))
		              {
			            case 0: return " Unavailable statistics.";
			            case 1: iStatisticSuccess = iStatisticValue;
					            break;
			            case 2: iStatisticAdapterTestFailure = iStatisticValue;
					            break;
			            case 3: iStatisticInsertionTestFailure = iStatisticValue;
					            break;
			            case 4: iStatisticIDCheckFailure = iStatisticValue;
					            break;
			            case 5: iStatisticAccessCanceledByUserFailure = iStatisticValue;
					            break;
			            case 6: iStatisticOtherFailure = iStatisticValue;
					            break;
			            default:iStatisticOperationalFail = iStatisticOperationalFail + iStatisticValue;
                                break;
		              }
		            }

	              }
	        } while (text != "");

	        return " Success:" + iStatisticSuccess.ToString() +
                   "  Oper. failure(s):" + iStatisticOperationalFail.ToString() +
		           "  Other failure(s):" + (iStatisticAdapterTestFailure +
											iStatisticInsertionTestFailure +
											iStatisticIDCheckFailure +
											iStatisticAccessCanceledByUserFailure +
                                            iStatisticOtherFailure).ToString() +
		           "  Total:" + (iStatisticSuccess +
								 iStatisticAdapterTestFailure +
								 iStatisticInsertionTestFailure +
								 iStatisticIDCheckFailure +
								 iStatisticAccessCanceledByUserFailure +
								 iStatisticOtherFailure +
                                 iStatisticOperationalFail).ToString();
        }

        // handler logic
        private void HandlerLogicRealization()
        {
            int iResultCode;
            string sCommandAnswer = "";

            if (!bSemaphore)
                return;

            bSemaphore = false;
            bool bCorrectFinish = false;

	        // reset result flag for all sites
	        // operation over each site of programmer
            for (int i = 0; i < iNumberSites; i++)
                // CHECK RESULT OF SITE
                if (!SendCommand(TXT_CMD_IDX_SITE_RESULT, i, ref sCommandAnswer))
                    return;

            try
            {
                // go to new device position
                if (!MovePusher(iNumberSites + (int)EnumSpriteElements.tSEReelIn, iNumberSites)) return;

                // signalize request for initial loading of devices to sockets
                bool bInitialSocketsLoading = true;

                int LoadDevice = 0;
      
                // infinity main loop
                while ((GUIControlState == EnumGUIControlState.tGCSBeforeStart) || (!SocketsEmpty()))
                {            
                    Application.DoEvents();
                    
                    // operation over each site of programmer
                    for (int i = 0; i < iNumberSites; i++)
                    {
                        // CHECK IF SITE IS PRESENT
                        if (!SendCommand(TXT_CMD_IDX_SITE_PRESENT, i, ref sCommandAnswer))
                            return;
                        if (sCommandAnswer != TXT_TRUE)
                        {
                            DisableSocket(i);

                            continue;
                        }
                        // site is present


                        EnableSocket(i);

                        // CHECK IF SITE IS BUSY
                        if (!SendCommand(TXT_CMD_IDX_SITE_BUSY, i, ref sCommandAnswer))
                            return;
                        if (sCommandAnswer != TXT_FALSE)
                        {
                            LedBusy(i, true); LedGood(i, false); LedError(i, false);

                            continue;
                        }
                        // site is not busy
                        LedBusy(i, false);


                        // CHECK RESULT OF SITE
                        if (!SendCommand(TXT_CMD_IDX_SITE_RESULT, i, ref sCommandAnswer))
                            return;

                        try
                        {
                            iResultCode = Convert.ToInt32(sCommandAnswer);
                        }
                        catch (FormatException)
                        {
                            iResultCode = 0;
                        }


                        // load device
                        if (bFullManualMode)
                            if ((GUIControlState == EnumGUIControlState.tGCSBeforeStart) && (LoadDevice + 1 <= iNumberSites))
                            {
                                LoadDevice++;
                                iResultCode = 0;
                            }

                        if (bInitialSocketsLoading)
                        {
                            // site result ... Unavailable
                            if (iResultCode != 0)
                                continue;
                        }
                        else // if (bInitialSocketsLoading)
                        {
                            // INCREMENT STATISTICS, REMOVE DEVICE FROM SOCKET AND INSERT IT TO OUTPUT REEL
                            // site result ... Unavailable
                            if (iResultCode == 0)
                                continue;

                            switch (GetActuatorStatus(i))
                            {
                                case 0: break;
                                case 1: break;
                                default: /*continue;*/ break;
                            }

                            if (comboBoxMode.SelectedIndex == 1)
                            {
                                FormOperationalResult formOperationalResult = new FormOperationalResult();
                                Point pointTemp = new Point();

                                pointTemp.X = panelControl.Left + 1;
                                pointTemp.Y = panelControl.Top + 1;
                                pointTemp = this.PointToScreen(pointTemp);
                                formOperationalResult.Left = pointTemp.X;
                                formOperationalResult.Top = pointTemp.Y;

                                if (!MovePusher(i, iNumberSites)) return;

                                formOperationalResult.SetButtonFocus(iResultCode == 1);
                                if (formOperationalResult.ShowDialog(this) == DialogResult.Yes)
                                    iResultCode = 1; // Good
                                else
                                    iResultCode = 7; // Operational Fail

                                formOperationalResult.Dispose();
                            }

                            IncrementSocketStatistic(i, iResultCode);
                            switch (iResultCode)
                            {
                                // site result ... Good
                                case 1: LedGood(i, true); LedError(i, false); break;
                                // site result ... Error
                                default: LedGood(i, false); LedError(i, true); break;
                            }

                            // get statistics for site to log
                            if (!SendCommand(TXT_CMD_IDX_SITE_STATISTICS, i, ref sCommandAnswer))
                                return;
                            else
                                WriteToLog("Statistics info for Site#" + (i + 1).ToString() + ReportStatistics(sCommandAnswer));

                            if (GetSocketState(i) == EnumSocketState.tSEnabledFull)
                            {
                                if (!RemoveSocketSequence(i)) return;

                                switch (iResultCode)
                                {
                                    // site result ... Good
                                    case 1: if (!ReelOperationSequence(EnumSpriteElements.tSEReelOutOK)) 
                                                return;
                                            break;
                                    // site result ... Error
                                    default: if (!ReelOperationSequence(EnumSpriteElements.tSEReelOutBad))
                                                return;
                                             break;
                                }
                            }
                            else
                                return;
                        } // else // if (bInitialSocketsLoading)


                        // TAKE NEW DEVICE, INSERT IT TO SOCKET AND START SITE
                        if (GUIControlState == EnumGUIControlState.tGCSBeforeStart)
                        {
                            if (!(GetSocketState(i) == EnumSocketState.tSEnabledEmpty)) return;

                            if (!ReelOperationSequence(EnumSpriteElements.tSEReelIn)) return;

                            if (!InsertSocketSequence(i)) return;

                            // start site #
                            if (!SendCommand(TXT_CMD_IDX_SITE_START, i, ref sCommandAnswer))
                                return;
                            GetActuatorStatus(i);            
                            if (sCommandAnswer == TXT_FALSE)
                                continue;

                            // CHECK IF SITE IS BUSY (Busy Led correction)
                            if (!SendCommand(TXT_CMD_IDX_SITE_BUSY, i, ref sCommandAnswer))
                                return;
                            if (sCommandAnswer == TXT_FALSE)
                            {
                                LedBusy(i, true); LedGood(i, false); LedError(i, false);
                            }
                        }
                    }  // for (int i = 0; i < iNumberSites; i++)
                    bInitialSocketsLoading = false;
                } // while ((GUIControlState == EnumGUIControlState.tGCSBeforeStart) || (!SocketsEmpty()))

                // go to new device position
                if (!MovePusher(iNumberSites + (int)EnumSpriteElements.tSEReelIn, iNumberSites)) return;

                // get total statistics to log
                if (!SendCommand(TXT_CMD_IDX_SITES_STATISTICS, ref sCommandAnswer))
                    return;
                else
                    WriteToLog("Summary:" + ReportStatistics(sCommandAnswer));

                bCorrectFinish = true;
            }
            finally
            {
                if (!bCorrectFinish)
                    WriteToLog("Error: Incorrect end of operation, analyse situation and continue by (Re)Initialization (recommended).");

                if (OnEventHandlerLogicRealizationFinished != null)
                    OnEventHandlerLogicRealizationFinished(bCorrectFinish);
                bSemaphore = true;
            }
        }

        // communication issues
        private bool SendCommand(string command, ref string answer, bool FromCmdLine = false)
        {
            if (uniHandlerClient == null)
            {
                WriteToLog("Command: " + command + " not accepted, (Re)Initialization is required.");
                return false;
            }

            if (uniHandlerClient.Write(command))
                if (uniHandlerClient.Read(ref answer))
                {
                    WriteToLog(command + " ... " + answer);

                    // successful finish
                    return true;
                }
                else
                {
                    if (answer != "")
                        WriteToLog(command + " ... " + answer);
                }

            if (!FromCmdLine)
                // incorrect finish
                SetGUIControlState(EnumGUIControlState.tGCSBeforeInitialization);

            return false;
        }
        private bool SendCommand(string command, int site, ref string answer)
        {
            const string TXT_SPACE = " ";

            return SendCommand(command + TXT_SPACE + (site + 1).ToString(), ref answer);
        }

        //// high level operation (Sequences)
        // remove (insert) device from (to) input (output) reel by pusher
        private bool ReelOperationSequence(EnumSpriteElements spriteElement)
        {          
            switch (spriteElement)
            {
                case EnumSpriteElements.tSEReelIn:
                case EnumSpriteElements.tSEReelOutOK:
                case EnumSpriteElements.tSEReelOutBad: break;
                default: return false;
            }

            if (!MovePusher(iNumberSites + (int)spriteElement, iNumberSites)) return false;

            switch (spriteElement)
            {
                case EnumSpriteElements.tSEReelIn:  if (!PullDeviceByPusher()) return false;
                                                    break;
                case EnumSpriteElements.tSEReelOutOK:
                case EnumSpriteElements.tSEReelOutBad:  if (!DropDeviceByPusher()) return false;
                                                        break;
            }

            if (!MoveReel(spriteElement)) return false;
            IncrementReelStatistic(spriteElement);

            return true;
        }
        // insert device to socket by pusher
        private bool InsertSocketSequence(int index)
        {
            // go to Index site
            if (!MovePusher(index, iNumberSites)) return false;
            if (!InsertDeviceToSocket(index)) return false;
            if (!DropDeviceByPusher()) return false;

            return true;
        }
        // remove device from socket by pusher
        private bool RemoveSocketSequence(int index)
        {
            // go to Index site
            if (!MovePusher(index, iNumberSites)) return false;
            if (!RemoveDeviceFromSocket(index)) return false;
            if (!PullDeviceByPusher()) return false;

            return true;
        }

        //// low level operation
        private Point GetPosition(int index, int aNumberSites)
        {
            switch ((EnumSpriteElements)(index - aNumberSites))
            {
                case EnumSpriteElements.tSEReelIn:
                case EnumSpriteElements.tSEReelOutOK:
                case EnumSpriteElements.tSEReelOutBad: return (spriteList[index] as ReelSprite).GetCentrePosition((spriteList[aNumberSites + (int)(EnumSpriteElements.tSEPusher)] as PusherSprite).GetWidth());
                default: return (spriteList[index] as Sprite).GetPosition();
            }      
        }
        private void PaintSprites()
        {
            // draw background            
            Graphics graphics = Graphics.FromImage(bitmapMerged);
            graphics.FillRectangle(new SolidBrush(SystemColors.ButtonFace), 0, 0, bitmapMerged.Size.Width, bitmapMerged.Size.Height);
            // draw sprites
            for (int i = 0; i < spriteList.Count; i++)
                (spriteList[i] as Sprite).Paint(ref graphics);
            pictureBox.Image = bitmapMerged;
            graphics.Dispose();
        }
        // move pusher from actual position to required NewPosition
        private bool MovePusher(int newPosition, int aNumberSites)
        {
            int iCounter, positionX, positionY, iDelta, iDeltaAxisX, iDeltaAxisY, iDirectionAxisX, iDirectionAxisY;
            Point pointFrom, pointTo;

            if (spriteList.Count <= 0)
                return false;

            if (!dedicatedDriver.MovePusher(newPosition, iNumberSites)) return false;

            if (newPosition != iCurrentPosition)
            {
                pointFrom = GetPosition(iCurrentPosition, aNumberSites);
                pointTo   = GetPosition(newPosition, aNumberSites);

                if (iStepSize != 5)
                {
                    // x,y axis movement
                    iDeltaAxisX = Math.Abs(pointFrom.X - pointTo.X);
                    iDeltaAxisY = Math.Abs(pointFrom.Y - pointTo.Y);
                    iDelta = Math.Max(iDeltaAxisX, iDeltaAxisY);
                    if ((pointTo.X - pointFrom.X) > 0)
                        iDirectionAxisX = 1;
                    else
                        iDirectionAxisX = -1;
                    if ((pointTo.Y - pointFrom.Y) > 0)
                        iDirectionAxisY = 1;
                    else
                        iDirectionAxisY = -1;

                    iCounter = 0;
                    while (iCounter < iDelta)
                    {
                        Application.DoEvents();
                        iCounter += iStepSize;
                        if (iDelta == iDeltaAxisX)
                            positionX = pointFrom.X + (iCounter * iDirectionAxisX);
                        else
                            positionX = pointFrom.X + ((iCounter * iDirectionAxisX * iDeltaAxisX) / iDelta);
                        if (iDelta == iDeltaAxisY)
                            positionY = pointFrom.Y + (iCounter * iDirectionAxisY);
                        else
                            positionY = pointFrom.Y + ((iCounter * iDirectionAxisY * iDeltaAxisY) / iDelta);

                        // set pusher position
                        (spriteList[aNumberSites + (int)EnumSpriteElements.tSEPusher] as Sprite).SetPosition(positionX, positionY);
                        PaintSprites();

                        if (iSleepSize > 0)
                            Thread.Sleep(iSleepSize);
                    }
                }
                else                
                    Application.DoEvents();

                // set pusher position
                (spriteList[aNumberSites + (int)EnumSpriteElements.tSEPusher] as Sprite).SetPosition(pointTo.X, pointTo.Y);
                PaintSprites();
                this.Refresh();

                if (iStepSize != 5)
                    if (iSleepSize > 0)
                        Thread.Sleep(250 * iSleepSize);
            }

            iCurrentPosition = newPosition;

            return true;

        }
        // pull device by pusher
        private bool PullDeviceByPusher()
        {
            if (!dedicatedDriver.PullDeviceByPusher())
                return false;

            if ((spriteList[iNumberSites + (int)EnumSpriteElements.tSEPusher] as PusherSprite).PusherState == EnumPusherState.tPEmpty)
            {
                (spriteList[iNumberSites + (int)EnumSpriteElements.tSEPusher] as PusherSprite).PusherState = EnumPusherState.tPFull;

                Graphics graphics = Graphics.FromImage(bitmapMerged);
                (spriteList[iNumberSites + (int)EnumSpriteElements.tSEPusher] as PusherSprite).Paint(ref graphics);
                pictureBox.Image = bitmapMerged;
                graphics.Dispose();
                return true;
            }
            return false;
        }
        // drop device by pusher
        private bool DropDeviceByPusher()
        {
            if (!dedicatedDriver.DropDeviceByPusher())
                return false;

            if ((spriteList[iNumberSites + (int)EnumSpriteElements.tSEPusher] as PusherSprite).PusherState == EnumPusherState.tPFull)
            {
                (spriteList[iNumberSites + (int)EnumSpriteElements.tSEPusher] as PusherSprite).PusherState = EnumPusherState.tPEmpty;

                Graphics graphics = Graphics.FromImage(bitmapMerged);
                (spriteList[iNumberSites + (int)EnumSpriteElements.tSEPusher] as PusherSprite).Paint(ref graphics);
                pictureBox.Image = bitmapMerged;
                graphics.Dispose();
                return true;
            }
            return false;
        }
        // insert device to socket
        private bool InsertDeviceToSocket(int index)
        {
            if (!dedicatedDriver.InsertDeviceToSocket(index))
                return false;

            if ((spriteList[index] as SocketSprite).SocketState == EnumSocketState.tSEnabledEmpty)
            {
                (spriteList[index] as SocketSprite).SocketState = EnumSocketState.tSEnabledFull;
                PaintSocket(index);
                return true;
            }
            return false;
        }
        // remove device from socket
        private bool RemoveDeviceFromSocket(int index)
        {
            if (!dedicatedDriver.RemoveDeviceFromSocket(index))
                return false;

            if ((spriteList[index] as SocketSprite).SocketState == EnumSocketState.tSEnabledFull)
            {
                (spriteList[index] as SocketSprite).SocketState = EnumSocketState.tSEnabledEmpty;
                PaintSocket(index);
                return true;
            }
            return false;
        }
        private int GetActuatorStatus(int index, bool bPaint = true)
        {
            string textTemp;

            int Result = -1;

            textTemp = "";
            if (!SendCommand(TXT_CMD_IDX_SITE_ACTUATOR_STATUS, index, ref textTemp))
                return Result;

            try
            {
                Result = Convert.ToInt32(textTemp);                
            }
            catch (FormatException)
            {
                Result = 3;
            }

            switch (Result)
            {
                case 0:  // 0 ... Error actuator unit is not present
                    (spriteList[index] as SocketSprite).SocketSideState = EnumSocketSideState.tSSideHidden;
                    break;
                case 1:  // 1 ... ZIF socket is opened by actuator
                    (spriteList[index] as SocketSprite).SocketSideState = EnumSocketSideState.tSSideEnabledDown;
                    break;
                case 2:  // 2 ... ZIF socket is closed by actuator
                    (spriteList[index] as SocketSprite).SocketSideState = EnumSocketSideState.tSSideEnabledUp;
                    break;
                default:
                    //   3 ... Communication error with actuator
                    //   4 ... ZIF socket actuator is in unknown position, please re-configure
                    (spriteList[index] as SocketSprite).SocketSideState = EnumSocketSideState.tSSideDisabledUp;
                    break;
            }

            if (bPaint)
                PaintSocket(index);

            return Result;
        }
    }
}
