﻿/*
	filename - UniHandlerClient.cs

	Purpose:
			 This file contains the declarations for implementation of Pipes (a section of shared memory that processes use for communication).
*/
/*
History (dd.mm.yy):
1.10.2010 start of development
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.IO.Pipes;

namespace VUH
{
    public class UniHandlerClient
    {
        public const string TXT_NACK = "NACK";
        public const string TXT_NACK_DELIMITER = " ";
        public const string TXT_NACK_ERROR_1 = "1";   // error .. communication error (wrong check sum, wrong packet)

        private NamedPipeClientStream hPipe = null;
        public delegate void LogProcedure(string text); // callback function for logging in higher-level layer
        private LogProcedure logProcedure;

        public UniHandlerClient(LogProcedure logProcedureArg)
        {
            logProcedure = new UniHandlerClient.LogProcedure(logProcedureArg);
        }
        ~UniHandlerClient()
        {            
        }

        private const int PACKET_PROTECTION_SIZE = 5; // Leading byte (1B), Data length (2B), Check sum (2B)
        private const int PACKET_SIZE = ushort.MaxValue + PACKET_PROTECTION_SIZE;
        private const byte PACKET_LEADING_BYTE = 0x7E;

        public bool Connect()
        {
            const string TXT_PIPE_NAME = "PG4UWMCPipeName4RemoteControl";

            try
            {
                hPipe = new NamedPipeClientStream(".", TXT_PIPE_NAME, PipeDirection.InOut);
                hPipe.Connect(1000);                

                if (!hPipe.IsConnected)
                    return false;
            }
            catch (System.Exception e)
            {
                ErrorMessageWindow(e.Message);
                return false;
            }
            return true;
        }
        public bool Disconnect()
        {
            try
            {
                hPipe.Close();
            }
            catch (System.Exception e)
            {
                ErrorMessageWindow(e.Message);
                return false;
            }
            return true;
        }
        private ushort CountCheckSum(byte[] data, int index, ushort length)
        {
            ushort result = 0;

            for (int i = 0; i < length; i++)
                result += (ushort)data[index + i];
            return result;
        }
        private bool WrapDataToPacket(ref byte[] packetData, string buffer)
        {            
            // check capability for PacketData length range
            int iLength = buffer.Length;
            if (iLength > ushort.MaxValue)
               return false;

            // set leading byte
            packetData[0] = PACKET_LEADING_BYTE;

            // set data length
            packetData[1] = (byte)(iLength & 0xFF);
            packetData[2] = (byte)(iLength >> 8);

            // copy data of packet
            for (int i = 0; i < iLength; i++)
              packetData[3 + i] = (byte)(buffer[i]);

            // count check sum
            ushort checkSum = CountCheckSum(packetData, 3, (ushort)iLength);

            // set check sum
            packetData[3 + iLength] = (byte)(checkSum & 0xFF);
            packetData[3 + iLength + 1] = (byte)(checkSum >> 8);

            return true;
        }
        private bool UnwrapDataFromPacket(ref string buffer, byte[] packetData)
        {
            buffer = "";

            // check packet integrity
            if (packetData[0] != PACKET_LEADING_BYTE)
                return false;

            int iLength = (packetData[2] << 8) + packetData[1];
            ushort checkSumFromPacket = (ushort)((packetData[3 + iLength + 1] << 8) + packetData[3 + iLength]);
            // compare packet check sum with calculated
            if (checkSumFromPacket != CountCheckSum(packetData, 3, (ushort)iLength))
                return false;

            // copy data
            for (int i = 0; i < iLength; i++)
                buffer = buffer + (char)packetData[3 + i];

            return true;
        }
        public bool Write(string buffer)
        {
            byte[] packetData = new byte[PACKET_SIZE];

            if (!WrapDataToPacket(ref packetData, buffer))
                return false;

            try
            {
                // write packet
                hPipe.Write(packetData, 0, PACKET_PROTECTION_SIZE + buffer.Length);
            }
            catch(System.Exception e)
            {
                ErrorMessageWindow(e.Message);
                return false;
            }
            return true;
        }
        public bool Read(ref string buffer)
        {
            byte[] packetData = new byte[PACKET_SIZE];
            buffer = "";

            try
            {
                // read packet
                hPipe.Read(packetData, 0, PACKET_SIZE);
            }
            catch (System.Exception e)
            {
                ErrorMessageWindow(e.Message);
                return false;
            }

            if (!UnwrapDataFromPacket(ref buffer, packetData))
            {
                buffer = TXT_NACK + TXT_NACK_DELIMITER + TXT_NACK_ERROR_1;
                return false;
            }

            return true;
        }
        private void WriteToLog(string text)
        {
            logProcedure(text);
        }
        private void ErrorMessageWindow(string text)
        {
            WriteToLog("Pipe Error: " + text);
        }
    }
}
